/*
 *name: MDialog.cpp
 *description: read $AXIS_ACT from OfficeLite
 *author: tilman jochems
 *(c)31/05/2010, mfj, rivesaltes
 */
#include <vcl.h>
#include <stdio.h>
#include <math.h>
#pragma hdrstop

#include "MDialog.h"

#define PORT               4849     //port of virtual arm server
#define COPYRIGHT "(c)2010, Midi-Formatique"

#pragma package(smart_init)
#pragma link "CrossCommEXE_OCX"
#pragma resource "*.dfm"
TFMDialog *FMDialog;

//construction
__fastcall TFMDialog::TFMDialog(TComponent* Owner)
        : TForm(Owner)
{
 WSADATA versioninfo;

 //position: right from kuka interface
 Left=min(640,Screen->Width-FMDialog->Width);Top=0;

 //initialize global variables
 Timeout=3000;                          //timeout in ms
 Position.SetLength(128);               //is this realy neccessary?
 Socket=INVALID_SOCKET;
 AddressOk=false;
 memset(&Address,0,sizeof(Address));
 Address.sin_family=AF_INET;
 Address.sin_port=htons(PORT);

#if DEBUG==1
 //simulation of crosscommexe
 Cross=new TCrossDebug();
#endif

#if DEBUG==0
 //init ole interface
 Cross->Connect();
 Cross->Init(NULL);
 Dispatch=NULL;
 Cross->Init(&Dispatch);
#endif

 //create UDP socket
 if (WSAStartup(MAKEWORD(2,0),&versioninfo)) throw("WSAStartup");
 if((Socket=socket(AF_INET,SOCK_DGRAM,0))==INVALID_SOCKET)
  throw Exception("create socket");
}//TFMDialog

//when coming up
void __fastcall TFMDialog::FormShow(TObject *Sender)
{
 int i1;

 //say hello
 MessageWrite("Hello, have fun!");

 //set address of arm server to localhost
 BSetClick(NULL);

 //get labels of axis
 for(i1=0;i1<6;i1++) LA[i1]=(TLabel *)FindComponent(AnsiString("LA")+i1);

 //set the idle handler
 Application->OnIdle=IdleHandler;
}//FormShow

//this is the end
void __fastcall TFMDialog::FormHide(TObject *Sender)
{
 //end communication with arm
 if(Socket!=INVALID_SOCKET){
  shutdown(Socket,SD_BOTH);
  closesocket(Socket);
 }//if
 WSACleanup();

 //free resources of roboter
 Cross->ServerOff();
}//FormHide

//called when application is idle idle
void __fastcall TFMDialog::IdleHandler(TObject* Sender,bool &Done){
 static ulong last=0;
 bool b1;
 int i1;
 ulong ul1;
 float a[6];

 //some states...
 static enum {TIMER_CONNECT,EXEC_CONNECT,EXEC_KRCOK,
              TIMER_SHOWVAR,EXEC_SHOWVAR,
              EXEC_SENDPOS} state=TIMER_CONNECT;

 //allways recall
 Done=false;

 //connect correctly to officelite
 switch(state){

  //connect to cross with frequency "timeout"
  case TIMER_CONNECT:
   if(GetTickCount()-last < (ulong)Timeout){
    Sleep(10);
    break;
   }//if
   last=GetTickCount();
   state=EXEC_CONNECT;
  break;

  case EXEC_CONNECT:                            //now connect
   if(!(bool)Cross->ConnectToCross(WideString("MFJPosServer"),0)){
    state=TIMER_CONNECT;                        //again and again
    break;
   }//if
   PCross->Color=clYellow;
   state=EXEC_KRCOK;
  break;

  //test if krc is ok (waits "timeout")
  case EXEC_KRCOK:
   if(!(bool)Cross->KrcOk(&Timeout)) break;     //wait until ok
   PCross->Color=clGreen;
   last=GetTickCount()-Timeout;                 //don't be too fast
   state=TIMER_SHOWVAR;
  break;

  //get position of axis all 200ms
  case TIMER_SHOWVAR:
   if(GetTickCount()-last < 200){
    Sleep(10);
    break;
   }//if
   last=GetTickCount();
   state=EXEC_SHOWVAR;
  break;

  case EXEC_SHOWVAR:                            //now read  position
   if(!(bool)Cross->ShowVar(WideString("$AXIS_ACT"),&Position,&Timeout)){
    FatalError("can't get variable");           //should only happen on
    break;                                      //end of officelite
   }//if
   state=EXEC_SENDPOS;
  break;

  //send position to arm server
  case EXEC_SENDPOS:
   //transform position to floats
   swscanf(Position,L"%*s = {%*s A1 %f, A2 %f, A3 %f, A4 %f, A5 %f, A6 %f",
           &a[0],&a[1],&a[2],&a[3],&a[4],&a[5],&a[6]);
   for(i1=0;i1<6;i1++) LA[i1]->Caption=AnsiString::FormatFloat("0.00",a[i1]);

   //send position to virtual arm
   if(AddressOk){
    sprintf(Buf,"%f %f %f %f %f %f",a[0],a[1],a[2],a[3],a[4],a[5],a[6]);
    i1=sendto(Socket,Buf,strlen(Buf),0,(sockaddr *)&Address,sizeof Address);
   }//if

   //show result of sending
   PArm->Color=(!AddressOk || i1==SOCKET_ERROR) ? clRed : clGreen;
   state=TIMER_SHOWVAR;
  break;

  default:
   FatalError("state not defined");
  break;
 }//switch
}//IdleHandler

//set address of arm server
void __fastcall TFMDialog::BSetClick(TObject *Sender)
{
 struct hostent *host;

 //get name of server
 host=gethostbyname(EAddress->Text.c_str());
 if(host==NULL){

  MessageWrite("IP address not valid",true);

  AddressOk=false;

  return;

 }//if


 //set address structure of server
 memcpy((char *)&Address.sin_addr,host->h_addr,sizeof Address.sin_addr);
 AddressOk=true;
}//BSetClick

//set address also on CR
void __fastcall TFMDialog::EAddressKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
 if(Key==VK_RETURN)BSet->Click();
}//EAdressKeyDown

//show messages on gui
void TFMDialog::MessageWrite(AnsiString msg,bool warn){
 LMessage->Caption=msg;
 LMessage->Color= warn ? clRed : clBtnFace;
 TMessage->Enabled=true;
}//MessageWrite

//end message
void __fastcall TFMDialog::TMessageTimer(TObject *Sender)
{
 TMessage->Enabled=false;
 LMessage->Color=clBtnFace;
 LMessage->Caption=COPYRIGHT;
}//TMessageTimer

//fatal error has occured -> close application
void TFMDialog::FatalError(AnsiString msg){
 MessageWrite(msg,true);
 Application->OnIdle=NULL;                      //stop idle handler
 Application->ProcessMessages();                //show message
 Sleep(5000);
 Close();                                       //that's it
}//FatalError
