/*
 *name: MDialog.cpp
 *description: read $AXIS_ACT from OfficeLite and send to vitual arm
 *author: tilman jochems
 *(c)06/06/2010, mfj, rivesaltes
 */
#include <vcl.h>
#include <syncobjs.hpp>
#include <stdio.h>
#include <math.h>
#pragma hdrstop

#include "MDialog.h"

#define PORT               4849     //port of virtual arm server
#define COPYRIGHT "(c)2010, Midi-Formatique"

#pragma package(smart_init)
#pragma link "CrossLib_OCX"
#pragma resource "*.dfm"
TFMDialog *FMDialog;

//construction
__fastcall TFMDialog::TFMDialog(TComponent* Owner)
        : TForm(Owner)
{
 WSADATA versioninfo;

 //position: right from kuka interface
 Left=min(640,Screen->Width-FMDialog->Width);Top=0;

 //initialize global variables
 SleepStart=false;                      //officelite starting?
 Timeout=3000;                          //timeout in ms
 Socket=INVALID_SOCKET;                 //socket+address of socket
 AddressOk=false;
 memset(&Address,0,sizeof(Address));
 Address.sin_family=AF_INET;
 Address.sin_port=htons(PORT);

 //create UDP socket
 if (WSAStartup(MAKEWORD(2,0),&versioninfo)) throw("WSAStartup");
 if((Socket=socket(AF_INET,SOCK_DGRAM,0))==INVALID_SOCKET)
  throw Exception("create socket");
}//TFMDialog

//when coming up
void __fastcall TFMDialog::FormShow(TObject *Sender)
{
 int i1;

 //say hello
 MessageWrite("Hello, have fun!");

 //set address of arm server to localhost
 BSetClick(NULL);

 //get labels of axis
 for(i1=0;i1<6;i1++) LA[i1]=(TLabel *)FindComponent(AnsiString("LA")+i1);

 //set the idle handler
 Application->OnIdle=IdleHandler;
}//FormShow

//this is the end
void __fastcall TFMDialog::FormHide(TObject *Sender)
{
 //end communication with arm
 if(Socket!=INVALID_SOCKET){
  shutdown(Socket,SD_BOTH);
  closesocket(Socket);
 }//if
 WSACleanup();

 //exit cross
 if(Cross!=NULL) Cross->Exit();
}//FormHide

//called when application is idle idle
void __fastcall TFMDialog::IdleHandler(TObject* Sender,bool &Done){
 static ulong last=0;
 int i1;
 float a[6];

 //some states...
 static enum {TIMER_CONNECT,EXEC_CONNECT,
              EXEC_KRCOK,RES_KRCOK,
              TIMER_FIRSTREAD,
              TIMER_SHOWVAR,EXEC_SHOWVAR,RES_SHOWVAR,
              EXEC_SENDPOS} state=TIMER_CONNECT;

 //allways recall
 Done=false;

 //some states in order to connect to officelite and to get the position
 switch(state){

  //connect to cross with frequency "timeout"
  case TIMER_CONNECT:
   PCross->Color=clRed;
   if(GetTickCount()-last < (ulong)Timeout){
    Sleep(20);
    break;
   }//if
   state=EXEC_CONNECT;
  break;

  case EXEC_CONNECT:
   if(!(bool)Cross->Init(WideString("MFJPosServerV2"),WideString(""),0)){
    state=TIMER_CONNECT;                        //again and again
    last=GetTickCount();
    break;
   }//if
   PCross->Color=clYellow;                      //we are connected
   state=EXEC_KRCOK;
  break;

  //wait for KRC to be ok
  case EXEC_KRCOK:
   JobOk=false;
   last=GetTickCount();
   Job=Cross->KRCOk();                          //start job
   Sleep(250);                                  //don't hurry, be happy
   state=RES_KRCOK;
  break;

  case RES_KRCOK:
   //test for timeout
   if(GetTickCount()-last > (ulong)Timeout){
    SleepStart=true;                            //this only happens if
    last=GetTickCount();                        //officelite starts
    state=TIMER_CONNECT;                        //connect again
    break;
   }//if

   //get result
   if(!JobOk){Sleep(20);break;}                 //wait for reponse from cross
   if(!(bool)Cross->KRCOkResult(Job)){
    state=EXEC_KRCOK;                           //try again
    break;
   }//if
   PCross->Color=clGreen;                       //we are ready
   last=GetTickCount();
   state=SleepStart ? TIMER_FIRSTREAD : TIMER_SHOWVAR;
  break;

  //after starting officelite, we have to sleep some time
  //why? (help from kuka would be appriciated)
  case TIMER_FIRSTREAD:
   if(GetTickCount()-last < 15000){
    Sleep(20);
    break;
   }//if
   SleepStart=false;                            //now, officelite should work
   state=TIMER_SHOWVAR;
  break;

  //get position of axis all 200ms
  case TIMER_SHOWVAR:
   if(GetTickCount()-last < 200){
    Sleep(20);
    break;
   }//if
   state=EXEC_SHOWVAR;
  break;

  case EXEC_SHOWVAR:
   JobOk=false;
   last=GetTickCount();                         //start job
   Job=Cross->ShowVar(WideString("$AXIS_ACT"),WideString(""));
   state=RES_SHOWVAR;
  break;

  case RES_SHOWVAR:
   //test for timeout
   if(GetTickCount()-last > (ulong)Timeout){
    FatalError("can't read variable");
    break;
   }//if

   //get result
   if(!JobOk){Sleep(1);break;}                 //wait for reponse from cross
   if(!(bool)Cross->ShowVarResult(Job,&Position)){
    state=TIMER_CONNECT;                        //connect again
    break;
   }//if
   state=EXEC_SENDPOS;
  break;

  //send position to arm server
  case EXEC_SENDPOS:
   //transform position to floats
   swscanf(Position,L"%*s = {%*s A1 %f, A2 %f, A3 %f, A4 %f, A5 %f, A6 %f",
           &a[0],&a[1],&a[2],&a[3],&a[4],&a[5],&a[6]);
   for(i1=0;i1<6;i1++) LA[i1]->Caption=AnsiString::FormatFloat("0.00",a[i1]);

   //send position to virtual arm
   if(AddressOk){
    sprintf(Buf,"%f %f %f %f %f %f",a[0],a[1],a[2],a[3],a[4],a[5],a[6]);
    i1=sendto(Socket,Buf,strlen(Buf),0,(sockaddr *)&Address,sizeof Address);
   }//if

   //show result of sending
   PArm->Color=(!AddressOk || i1==SOCKET_ERROR) ? clRed : clGreen;
   state=TIMER_SHOWVAR;
  break;
 }//switch
}//IdleHandler

//message send to application when job has finished
void __fastcall TFMDialog::CrossCmdMessage(TObject *Sender, short nJob,
      long lError)
{
 //if it is the actual job -> result can be read
 if(nJob==Job) JobOk=true;
}//CrossCmdMessage

//set address of arm server
void __fastcall TFMDialog::BSetClick(TObject *Sender)
{
 struct hostent *host;

 //get name of server
 host=gethostbyname(EAddress->Text.c_str());
 if(host==NULL){

  MessageWrite("IP address not valid",true);

  AddressOk=false;

  return;

 }//if


 //set address structure of server
 memcpy((char *)&Address.sin_addr,host->h_addr,sizeof Address.sin_addr);
 AddressOk=true;
}//BSetClick

//set address also on CR
void __fastcall TFMDialog::EAddressKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
 if(Key==VK_RETURN)BSet->Click();
}//EAdressKeyDown

//show messages on gui
void TFMDialog::MessageWrite(AnsiString msg,bool warn){
 LMessage->Caption=msg;
 LMessage->Color= warn ? clRed : clBtnFace;
 TMessage->Enabled=true;
}//MessageWrite

//end message
void __fastcall TFMDialog::TMessageTimer(TObject *Sender)
{
 TMessage->Enabled=false;
 LMessage->Color=clBtnFace;
 LMessage->Caption=COPYRIGHT;
}//TMessageTimer

//fatal error has occured -> close application
void TFMDialog::FatalError(AnsiString msg){
 delete Cross;                                  //stop ole
 Cross=NULL;
 MessageWrite(msg,true);
 Application->OnIdle=NULL;                      //stop idle handler
 Application->ProcessMessages();                //show message
 Sleep(2500);
 Close();                                       //that's it
}//FatalError

